/*
* Scilab ( http://www.scilab.org/ ) - This file is part of Scilab
* Copyright (C) 2014 - Scilab Enterprises - Cedric DELAMARRE
*
* This file must be used under the terms of the CeCILL.
* This source file is licensed as described in the file COPYING, which
* you should have received as part of this distribution.  The terms
* are also available at
* http://www.cecill.info/licences/Licence_CeCILL_V2-en.txt
*
*/

#include "config_gpu.h"
#ifdef WITH_Nvidia
#pragma OPENCL EXTENSION cl_khr_fp64: enable
#endif

double ddot_gpu(int N, __global double* DX, int INCX, __global double* DY, int INCY)
{
    int ix = 0;
    int iy = 0;

    if(INCX < 0)
    {
        ix = (-N + 1) * INCX;
    }

    if(INCY < 0)
    {
        iy = (-N + 1) * INCY;
    }

    double tmp = 0;
    for (int i = 0; i < N; i++)
    {
        tmp += DX[ix] * DY[iy];
        ix  += INCX;
        iy  += INCY;
    }

    return tmp;
}

__kernel void compute_conv(__global double* A, int mA, int nA,
                           __global double* B, int mB, int nB,
                           __global double* Out, int mOut, int nOut,
                           int edgM, int edgN)
{
    int idCols = get_global_id(0);
    int idRows = get_global_id(1);

    if(idRows >= mOut || idCols >= nOut)
    {
        return;
    }

    int bi = mB - 1 - max(0, edgM - idRows);
    int ai = max(0, idRows - edgM);
    int aj = 0;
    int bj = 0;

    double sum = 0;

    for (bj = nB - 1 - max(0, edgN - idCols), aj = max(0, idCols - edgN); bj >= 0 && aj < nA; bj--, aj++)
    {
        int l = min(bi + 1, mA - ai);
        sum += ddot_gpu(l, A + ai + mA * aj, 1, B + bi - l + 1 + mB * bj, -1);
    }

    Out[idRows + idCols * mOut] = sum;
}

__kernel void compute_copy(__global double* src, __global double* dest, int iRows, int iCols)
{
    int idCols = get_global_id(0);
    int idRows = get_global_id(1);

    if(idRows >= iRows || idCols >= iCols)
    {
        return;
    }

    int id = idCols * iRows + idRows;
    dest[id] = src[id];
}

__kernel void compute_vxvy(__global double* level, __global double* vx, __global double* vy, int iRows, int iCols)
{
    int idCols = get_global_id(0);
    int idRows = get_global_id(1);

    if(idRows >= iRows || idCols >= iCols)
    {
        return;
    }

    int id   = idCols * iRows + idRows;
    int up   = id + 1;
    int left = id + iRows;

    if(idRows == iRows - 1)
    {
        up = idCols * iRows;
    }

    if(idCols == iCols - 1)
    {
        left = idRows;
    }

    vx[id] *= 0.99998;
    vy[id] *= 0.99998;
    vx[id] += 0.0001 * (level[id] - level[left]);
    vy[id] += 0.0001 * (level[id] - level[up]);
}

__kernel void compute_level(__global double* level, __global double* vx, __global double* vy, int iRows, int iCols)
{
    int idCols = get_global_id(0);
    int idRows = get_global_id(1);

    if(idRows >= iRows || idCols >= iCols)
    {
        return;
    }

    int id    = idCols * iRows + idRows;
    int down  = id - 1;
    int right = id - iRows;

    if(idRows == 0)
    {
        down = id + iRows - 1;
    }

    if(idCols == 0)
    {
        right = iRows * (iCols - 1) + idRows;
    }

    level[id] -= vy[id] - vy[down] + vx[id] - vx[right];
}

