/*
* Scilab ( http://www.scilab.org/ ) - This file is part of Scilab
* Copyright (C) Scilab Enterprises - 2014 - Cedric DELAMARRE
*
* This file must be used under the terms of the CeCILL.
* This source file is licensed as described in the file COPYING, which
* you should have received as part of this distribution.  The terms
* are also available at
* http://www.cecill.info/licences/Licence_CeCILL_V2-en.txt
*
*/

#include "config_gpu.h"
#ifdef WITH_Nvidia
#pragma OPENCL EXTENSION cl_khr_fp64: enable
#endif


/*****************/
/*    A .*. B    */
/*****************/

// number of threads = size of B, kernel loop on A
__kernel void kroneckerB_kernel(__global double* dA, int iRowsA, int iColsA, int bAIsComplex,
                                __global double* dB, int iRowsB, int iColsB, int bBIsComplex,
                                __global double* dOut, int iRowsOut)
{
    int iPos = get_global_id(0);
    if(iPos >= iRowsB*iColsB)
    {
        return;
    }

    int iPosX = iPos / iRowsB;
    int iPosY = iPos % iRowsB;

    if(bAIsComplex == 0 && bBIsComplex == 0)
    {// real case
        for(int i = 0; i < iColsA; i++)
        {
            for(int j = 0; j < iRowsA; j++)
            {
                int pos = (i * iColsB + iPosX) * iRowsOut + j * iRowsB + iPosY;
                dOut[pos] = dA[i*iRowsA+j] * dB[iPos];
            }
        }
    }
    else if(bAIsComplex && bBIsComplex == 0)
    {// A complex
        __global double2* out = (__global double2*)dOut;
        __global double2* da  = (__global double2*)dA;
        for(int i = 0; i < iColsA; i++)
        {
            for(int j = 0; j < iRowsA; j++)
            {
                int iPosOut = (i * iColsB + iPosX) * iRowsOut + j * iRowsB + iPosY;
                int iPosA = i*iRowsA+j;
                out[iPosOut].x = da[iPosA].x * dB[iPos];
                out[iPosOut].y = da[iPosA].y * dB[iPos];
            }
        }
    }
    else if(bAIsComplex == 0 && bBIsComplex)
    {// B complex
        __global double2* out = (__global double2*)dOut;
        __global double2* db  = (__global double2*)dB;
        for(int i = 0; i < iColsA; i++)
        {
            for(int j = 0; j < iRowsA; j++)
            {
                int iPosOut = (i * iColsB + iPosX) * iRowsOut + j * iRowsB + iPosY;
                int iPosA = i*iRowsA+j;
                out[iPosOut].x = dA[iPosA] * db[iPos].x;
                out[iPosOut].y = dA[iPosA] * db[iPos].y;
            }
        }
    }
    else
    {// A and B complex
        __global double2* out = (__global double2*)dOut;
        __global double2* da  = (__global double2*)dA;
        __global double2* db  = (__global double2*)dB;
        for(int i = 0; i < iColsA; i++)
        {
            for(int j = 0; j < iRowsA; j++)
            {
                int iPosOut = (i * iColsB + iPosX) * iRowsOut + j * iRowsB + iPosY;
                int iPosA = i*iRowsA+j;
                out[iPosOut].x = da[iPosA].x * db[iPos].x - da[iPosA].y * db[iPos].y;
                out[iPosOut].y = da[iPosA].x * db[iPos].y + da[iPosA].y * db[iPos].x;
            }
        }
    }
}

// number of threads = size of A, kernel loop on B
__kernel void kroneckerA_kernel(__global double* dA, int iRowsA, int iColsA, int bAIsComplex,
                                __global double* dB, int iRowsB, int iColsB, int bBIsComplex,
                                __global double* dOut, int iRowsOut)
{
    int iPos = get_global_id(0);

    if(iPos >= iRowsA*iColsA)
    {
        return;
    }

    int iPosX = iPos / iRowsA;
    int iPosY = iPos % iRowsA;

    if(bAIsComplex == 0 && bBIsComplex == 0)
    {// real case
        for(int i = 0; i < iColsB; i++)
        {
            for(int j = 0; j < iRowsB; j++)
            {
                int iPosOut = (iPosX * iColsB + i) * iRowsOut + iPosY * iRowsB + j;
                dOut[iPosOut] = dA[iPos] * dB[i*iRowsB+j];
            }
        }
    }
    else if(bAIsComplex && bBIsComplex == 0)
    {// A complex
        __global double2* out = (__global double2*)dOut;
        __global double2* da  = (__global double2*)dA;
        for(int i = 0; i < iColsB; i++)
        {
            for(int j = 0; j < iRowsB; j++)
            {
                int iPosOut = (iPosX * iColsB + i) * iRowsOut + iPosY * iRowsB + j;
                int iPosB = i*iRowsB+j;
                out[iPosOut].x = da[iPos].x * dB[iPosB];
                out[iPosOut].y = da[iPos].y * dB[iPosB];
            }
        }
    }
    else if(bAIsComplex == 0 && bBIsComplex)
    {// B complex
        __global double2* out = (__global double2*)dOut;
        __global double2* db  = (__global double2*)dB;
        for(int i = 0; i < iColsB; i++)
        {
            for(int j = 0; j < iRowsB; j++)
            {
                int iPosOut = (iPosX * iColsB + i) * iRowsOut + iPosY * iRowsB + j;
                int iPosB = i*iRowsB+j;
                out[iPosOut].x = dA[iPos] * db[iPosB].x;
                out[iPosOut].y = dA[iPos] * db[iPosB].y;
            }
        }
    }
    else
    {// A and B complex
        __global double2* out = (__global double2*)dOut;
        __global double2* da  = (__global double2*)dA;
        __global double2* db  = (__global double2*)dB;
        for(int i = 0; i < iColsB; i++)
        {
            for(int j = 0; j < iRowsB; j++)
            {
                int iPosOut = (iPosX * iColsB + j) * iRowsOut + iPosY * iRowsB + i;
                int iPosB = i*iRowsB+j;
                out[iPosOut].x = da[iPos].x * db[iPosB].x - da[iPos].y * db[iPosB].y;
                out[iPosOut].y = da[iPos].x * db[iPosB].y + da[iPos].y * db[iPosB].x;
            }
        }
    }
}

// grid size = size of A, Block size = size of B
__kernel void kroneckerAB_kernel(__global double* dA, int bAIsComplex,
                                 __global double* dB, int bBIsComplex,
                                 __global double* dOut, int iRowsOut)
{
    int iPosX = get_global_id(0);
    int iPosY = get_global_id(1);

    int iPosA = get_group_id(0) * get_num_groups(1) + get_group_id(1);
    int iPosB = get_local_id(0) * get_local_size(1) + get_local_id(1);

    if(bAIsComplex == 0 && bBIsComplex == 0)
    {// real case
        dOut[iPosX * iRowsOut + iPosY] = dA[iPosA] * dB[iPosB];
    }
    else if(bAIsComplex && bBIsComplex == 0)
    {// A complex
        __global double2* out = (__global double2*)dOut;
        __global double2* da  = (__global double2*)dA;
        out[iPosX * iRowsOut + iPosY].x = da[iPosA].x * dB[iPosB];
        out[iPosX * iRowsOut + iPosY].y = da[iPosA].y * dB[iPosB];
    }
    else if(bAIsComplex == 0 && bBIsComplex)
    {// B complex
        __global double2* out = (__global double2*)dOut;
        __global double2* db  = (__global double2*)dB;
        out[iPosX * iRowsOut + iPosY].x = dA[iPosA] * db[iPosB].x;
        out[iPosX * iRowsOut + iPosY].y = dA[iPosA] * db[iPosB].y;
    }
    else
    {// A and B complex
        __global double2* out = (__global double2*)dOut;
        __global double2* da  = (__global double2*)dA;
        __global double2* db  = (__global double2*)dB;
        out[iPosX * iRowsOut + iPosY].x = da[iPosA].x * db[iPosB].x - da[iPosA].y * db[iPosB].y;
        out[iPosX * iRowsOut + iPosY].y = da[iPosA].x * db[iPosB].y + da[iPosA].y * db[iPosB].x;
    }
}
