/*
* Scilab ( http://www.scilab.org/ ) - This file is part of Scilab
* Copyright (C) Scilab Enterprises - 2014 - Cedric DELAMARRE
*
* This file must be used under the terms of the CeCILL.
* This source file is licensed as described in the file COPYING, which
* you should have received as part of this distribution.  The terms
* are also available at
* http://www.cecill.info/licences/Licence_CeCILL_V2-en.txt
*
*/

#include "config_gpu.h"
#ifdef WITH_Nvidia
#pragma OPENCL EXTENSION cl_khr_fp64: enable
#endif

__kernel void max_kernel(__global double* d, int elems, __global double* result, __local double* accumResult)
{
    int iGlobalPos      = get_global_id(0);
    int iLocalPos       = get_local_id(0);
    int iLocalSize      = get_local_size(0);

    if(iGlobalPos < elems)
    {
        accumResult[iLocalPos] = d[iGlobalPos];
    }

    barrier(CLK_GLOBAL_MEM_FENCE);

    // if iLocalSize is odd.
    if(iLocalSize & 1 && iLocalPos == 0)
    {
        if(accumResult[iLocalSize-1] > accumResult[0])
        {
            accumResult[0] = accumResult[iLocalSize-1];
        }
    }

    for(int stride = iLocalSize >> 1; stride > 0; stride >>= 1)
    {
        if(iLocalPos < stride && iGlobalPos + stride < elems)
        {
            if(accumResult[stride + iLocalPos] > accumResult[iLocalPos])
                accumResult[iLocalPos] = accumResult[stride + iLocalPos];
        }

        barrier(CLK_LOCAL_MEM_FENCE);

        // if stride is odd, compare last element with the first
        // ie: matrix = 1 2 3 4 5 6 7
        // compare 1 and 7
        // then next loop will compare 7 (old 1) and 4, 2 and 5, 3 and 6
        if(stride & 1 && iLocalPos == 0)
        {
            if(accumResult[stride-1] > accumResult[0])
            {
                accumResult[0] = accumResult[stride-1];
            }
        }

        barrier(CLK_LOCAL_MEM_FENCE);
    }

    if(iLocalPos == 0) result[get_group_id(0)] = accumResult[0];
}


__kernel void min_kernel(__global double* d, int elems, __global double* result, __local double* accumResult)
{
    int iGlobalPos      = get_global_id(0);
    int iLocalPos       = get_local_id(0);
    int iLocalSize      = get_local_size(0);

    if(iGlobalPos < elems)
    {
        accumResult[iLocalPos] = d[iGlobalPos];
    }

    barrier(CLK_GLOBAL_MEM_FENCE);

    // if iLocalSize is odd.
    if(iLocalSize & 1 && iLocalPos == 0)
    {
        if(accumResult[iLocalSize-1] < accumResult[0])
        {
            accumResult[0] = accumResult[iLocalSize-1];
        }
    }

    for(int stride = iLocalSize >> 1; stride > 0; stride >>= 1)
    {
        if(iLocalPos < stride && iGlobalPos + stride < elems)
        {
            if(accumResult[stride + iLocalPos] < accumResult[iLocalPos])
                accumResult[iLocalPos] = accumResult[stride + iLocalPos];
        }

        barrier(CLK_LOCAL_MEM_FENCE);

        // if stride is odd, compare last element with the first
        // ie: matrix = 1 2 3 4 5 6 7
        // compare 1 and 7
        // then next loot will compare 7 (old 1) and 4, 2 and 5, 3 and 6
        if(stride & 1 && iLocalPos == 0)
        {
            if(accumResult[stride-1] < accumResult[0])
            {
                accumResult[0] = accumResult[stride-1];
            }
        }

        barrier(CLK_LOCAL_MEM_FENCE);
    }

    if(iLocalPos == 0) result[get_group_id(0)] = accumResult[0];
}

__kernel void elementWiseMax(__global double *odata, __global double* idataA, __global double* idataB, int width, int height)
{
    unsigned int xIndex = get_global_id(0);
    unsigned int yIndex = get_global_id(1);
    unsigned int index  = yIndex + height * xIndex;

    if (xIndex < width && yIndex < height)
    {
        if(idataA[index] < idataB[index])
        {
            odata[index] = idataB[index];
        }
        else
        {
            odata[index] = idataA[index];
        }
    }
}

__kernel void elementWiseZMax(__global double2* odata, __global double2* idataA, __global double2* idataB, int width, int height)
{
    unsigned int xIndex = get_global_id(0);
    unsigned int yIndex = get_global_id(1);
    unsigned int index  = yIndex + height * xIndex;
    double valA = 0;
    double valB = 0;

    if (xIndex < width && yIndex < height)
    {
        valA = fabs(idataA[index].x) + fabs(idataA[index].y);
        valB = fabs(idataB[index].x) + fabs(idataB[index].y);
        if(valA < valB)
        {
            odata[index].x = idataB[index].x;
            odata[index].y = idataB[index].y;
        }
        else
        {
            odata[index].x = idataA[index].x;
            odata[index].y = idataA[index].y;
        }
    }
}

__kernel void elementWiseZDMax(__global double2* odata, __global double2* idataA, __global double* idataB, int width, int height)
{
    unsigned int xIndex = get_global_id(0);
    unsigned int yIndex = get_global_id(1);
    unsigned int index  = yIndex + height * xIndex;
    double valA = 0;
    double valB = 0;

    if (xIndex < width && yIndex < height)
    {
        valA = fabs(idataA[index].x) + fabs(idataA[index].y);
        valB = fabs(idataB[index]);
        if(valA < valB)
        {
            odata[index].x = idataB[index];
            odata[index].y = 0;
        }
        else
        {
            odata[index].x = idataA[index].x;
            odata[index].y = idataA[index].y;
        }
    }
}

__kernel void elementWiseMin(__global double *odata, __global double* idataA, __global double* idataB, int width, int height)
{
    unsigned int xIndex = get_global_id(0);
    unsigned int yIndex = get_global_id(1);
    unsigned int index  = yIndex + height * xIndex;

    if (xIndex < width && yIndex < height)
    {
        if(idataA[index] > idataB[index])
        {
            odata[index] = idataB[index];
        }
        else
        {
            odata[index] = idataA[index];
        }
    }
}

__kernel void elementWiseZMin(__global double2* odata, __global double2* idataA, __global double2* idataB, int width, int height)
{
    unsigned int xIndex = get_global_id(0);
    unsigned int yIndex = get_global_id(1);
    unsigned int index  = yIndex + height * xIndex;
    double valA = 0;
    double valB = 0;

    if (xIndex < width && yIndex < height)
    {
        valA = fabs(idataA[index].x) + fabs(idataA[index].y);
        valB = fabs(idataB[index].x) + fabs(idataB[index].y);
        if(valA > valB)
        {
            odata[index].x = idataB[index].x;
            odata[index].y = idataB[index].y;
        }
        else
        {
            odata[index].x = idataA[index].x;
            odata[index].y = idataA[index].y;
        }
    }
}

__kernel void elementWiseZDMin(__global double2* odata, __global double2* idataA, __global double* idataB, int width, int height)
{
    unsigned int xIndex = get_global_id(0);
    unsigned int yIndex = get_global_id(1);
    unsigned int index  = yIndex + height * xIndex;
    double valA = 0;
    double valB = 0;

    if (xIndex < width && yIndex < height)
    {
        valA = fabs(idataA[index].x) + fabs(idataA[index].y);
        valB = fabs(idataB[index]);
        if(valA > valB)
        {
            odata[index].x = idataB[index];
            odata[index].y = 0;
        }
        else
        {
            odata[index].x = idataA[index].x;
            odata[index].y = idataA[index].y;
        }
    }
}
