/*
* Scilab ( http://www.scilab.org/ ) - This file is part of Scilab
* Copyright (C) Scilab Enterprises - 2014 - Cedric DELAMARRE
*
* This file must be used under the terms of the CeCILL.
* This source file is licensed as described in the file COPYING, which
* you should have received as part of this distribution.  The terms
* are also available at
* http://www.cecill.info/licences/Licence_CeCILL_V2-en.txt
*
*/

#include "config_gpu.h"
#ifdef WITH_Nvidia
#pragma OPENCL EXTENSION cl_khr_fp64: enable
#endif

#include "splinType.h"

/*****************************************/
/************ Device funtions ************/
/*****************************************/
// Sign-Testing Routine.
// Returns:
//    -1. if ARG1 and ARG2 are of opposite sign.
//     0. if either argument is zero.
//    +1. if ARG1 and ARG2 are of the same sign.
int dpchst(double arg1, double arg2)
{
    if(arg1 == 0 || arg2 == 0)
    {
        return 0;
    }

    return (arg1 * arg2) < 0 ? -1 : 1;
}
int computeCPos(int x, int y, int X, int Y, int sizeOfX)
{
    // C(4,4,sizeOfX-1,sizeOfY-1)
    // C(x,y,X,Y)
    return x + y*4 + 16*X + 16*(sizeOfX-1)*Y;
}
void triDiagLDLSolve(__global double* Ad, __global double* Asd, __global double* out, int sizeOfIn, int incr)
{
    for (int i = incr; i < sizeOfIn*incr; i+=incr)
    {
        double temp = Asd[i-incr];
        Asd[i-incr] = Asd[i-incr] / Ad[i-incr];
        Ad[i] = Ad[i] - temp * Asd[i-incr];
        out[i] = out[i] - Asd[i-incr] * out[i-incr];
    }

    out[(sizeOfIn-1)*incr] = out[(sizeOfIn-1)*incr] / Ad[(sizeOfIn-1)*incr];

    for (int i = (sizeOfIn-2)*incr; i >= 0; i-=incr)
    {
        out[i] = out[i] / Ad[i] - Asd[i] * out[i+incr];
    }
}
void cyclicTriDiagLDLtSolve(__global double* Ad, __global double* Asd, __global double* Qdu, __global double* out, int sizeOfIn, int incr)
{
    double temp1 = 0;
    double temp2 = 0;

    //compute the LDL^t factorization
    for (int i = 0; i < (sizeOfIn-2)*incr; i+=incr)
    {
        temp1 = Asd[i];
        temp2 = Qdu[i];
        Asd[i] = Asd[i] / Ad[i]; //elimination coef L(i,i-1)
        Qdu[i] = Qdu[i] / Ad[i]; //elimination coef L(i,i-1)

        Ad[i+incr] = Ad[i+incr] - Asd[i] * temp1; // elimination on line i+1
        Qdu[i+incr] = Qdu[i+incr] - Qdu[i] * temp1; // elimination on line n
        Ad[(sizeOfIn-1)*incr] = Ad[(sizeOfIn-1)*incr] - Qdu[i] * temp2; // elimination on line n
    }

    temp2 = Qdu[(sizeOfIn-2)*incr];
    Qdu[(sizeOfIn-2)*incr] = Qdu[(sizeOfIn-2)*incr] / Ad[(sizeOfIn-2)*incr];
    Ad[(sizeOfIn-1)*incr] = Ad[(sizeOfIn-1)*incr] - Qdu[(sizeOfIn-2)*incr] * temp2;

    // solve LDL^t x = b  (but use b for x and for the intermediary vectors...)
    for(int i = incr; i < (sizeOfIn-1)*incr; i+=incr)
    {
        out[i] = out[i] - Asd[i-incr] * out[i-incr];
    }

    for(int i = 0; i < (sizeOfIn-1)*incr; i+=incr)
    {
        out[(sizeOfIn-1) * incr] = out[(sizeOfIn-1) * incr] - Qdu[i] * out[i];
    }

    for(int i = 0; i < sizeOfIn*incr; i+=incr)
    {
        out[i] = out[i] / Ad[i];
    }

    out[(sizeOfIn-2) * incr] = out[(sizeOfIn-2) * incr] - Qdu[(sizeOfIn-2)*incr] * out[(sizeOfIn-1) * incr];
    for (int i = (sizeOfIn-3)*incr; i >= 0; i-=incr)
    {
        out[i] = out[i] - Asd[i] * out[i+incr] - Qdu[i]*out[(sizeOfIn-1) * incr];
    }
}

/**************************************/
/************ splin Kernel ************/
/**************************************/
__kernel void derivd_FP_kernel(int sizeOfIn, int iSize, __global double* In, __global double* Z,
                                __global double* Out, int incr)
{// Fast Periodic
    int iPosX = get_global_id(0);
    int iPosY = get_global_id(1);
    int iPos  = iPosY * sizeOfIn + iPosX;

    if(iPosX >= iSize || iPosY >= iSize )
    {
        return;
    }

    __global double* z = Z + iPos;
    __global double* out = Out + iPos;

    if(sizeOfIn == 2)
    {
        out[0] = (z[incr] - z[0]) / (In[1] - In[0]);
        out[incr] = out[0];
        return;
    }

    double dx_r = In[sizeOfIn-1] - In[sizeOfIn-2];
    double du_r = (z[0] - z[(sizeOfIn-2) * incr]) / dx_r;

    for (int i = 0; i < sizeOfIn - 1; i++)
    {
        double dx_l = dx_r;
        double du_l = du_r;
        dx_r = In[i+1] - In[i];
        du_r = (z[(i+1) * incr] - z[i*incr]) / dx_r;
        double w_l = dx_r / (dx_l + dx_r);
        double w_r = 1 - w_l;
        out[i*incr] = w_l * du_l + w_r * du_r;
    }

    out[(sizeOfIn-1) * incr] = out[0];
}

__kernel void derivd_F_kernel(int _iPos, int sizeOfIn, int iSize, __global double* In, __global double* Z,
                              __global double* Out, int incr)
{// Fast
    int iPosX = get_global_id(0);
    int iPosY = get_global_id(1);
    int iPos = _iPos;

    if(iPosX >= iSize || iPosY >= iSize )
    {
        return;
    }

    if(iPos == 0)
    {
        iPos = iPosY * sizeOfIn + iPosX;
    }

    __global double* z = Z + iPos;
    __global double* out = Out + iPos;

    if(sizeOfIn == 2)
    {
        out[0] = (z[incr] - z[0]) / (In[1] - In[0]);
        out[incr] = out[0];
        return;
    }

    double dx_l = In[1] - In[0];
    double du_l = (z[incr] - z[0]) / dx_l;
    double dx_r = In[2] - In[1];
    double du_r = (z[2*incr] - z[incr]) / dx_r;
    double w_l = dx_r / (dx_l + dx_r);
    double w_r = 1 - w_l;
    out[0] = (1 + w_r) * du_l - w_r * du_r;
    out[incr] = w_l * du_l + w_r * du_r;

    for (int i = 2; i < sizeOfIn - 1; i++)
    {
        dx_l = dx_r;
        du_l = du_r;
        dx_r = In[i+1] - In[i];
        du_r = (z[(i+1)*incr] - z[i*incr]) / dx_r;
        w_l = dx_r / (dx_l + dx_r);
        w_r = 1 - w_l;
        out[i*incr] = w_l * du_l + w_r * du_r;
    }

    out[(sizeOfIn-1) * incr] = (1 + w_l) * du_r - w_l * du_l;
}

__kernel void dpchim_kernel(int sizeOfIn, int iSize, __global double* In, __global double* Z,
                             __global double* Out, int incr)
{// Monotone
    int iPosX = get_global_id(0);
    int iPosY = get_global_id(1);
    int iPos  = iPosY * sizeOfIn + iPosX;

    if(iPosX >= iSize || iPosY >= iSize )
    {
        return;
    }

    __global double* z = Z + iPos;
    __global double* out = Out + iPos;

    double h1 = In[1] - In[0];
    double del1 = (z[incr] - z[0]) / h1;
    // case sizeOfIn = 2, use linear interpolation
    if(sizeOfIn == 2)
    {
        out[0] = del1;
        out[(sizeOfIn-1) * incr] = del1;
        return;
    }

    // normal case
    double h2 = In[2] - In[1];
    double del2 = (z[2 * incr] - z[incr]) / h2;

    // set out[0] via non-centered three-point formula,
    // adjusted to be shape-preserving
    double hsum = h1 + h2;
    double w1 = (h1 + hsum) / hsum;
    double w2 = -h1 / hsum;
    out[0] = w1 * del1 + w2 * del2;
    if (dpchst(out[0], del1) <= 0)
    {
        out[0] = 0;
    }
    else if(dpchst(del1, del2) < 0)
    {
        double dmax = 3 * del1;
        if(fabs(out[0]) > fabs(dmax))
        {
            out[0] = dmax;
        }
    }

    // loop through interior points
    for(int i = 1; i < sizeOfIn - 1; i++)
    {
        if (i != 1)
        {
            h1 = h2;
            h2 = In[i+1] - In[i];
            hsum = h1 + h2;
            del1 = del2;
            del2 = (z[(i+1)*incr] - z[i*incr]) / h2;
        }

        // set out[i] = 0 unless data are strictly monotonic
        out[i * incr] = 0;
        if(dpchst(del1, del2) > 0)
        {
            double hsumt3 = 3*hsum;
            w1 = (hsum + h1) / hsumt3;
            w2 = (hsum + h2) / hsumt3;
            double fabsdel1 = fabs(del1);
            double fabsdel2 = fabs(del2);
            double m = max(fabsdel1, fabsdel2);
            double drat1 = del1 / m;
            double drat2 = del2 / m;
            out[i*incr] = min(fabsdel1, fabsdel2) / (w1 * drat1 + w2 * drat2);
        }
    }

    w1 = -h2 / hsum;
    w2 = (h2 + hsum) / hsum;
    iPos = (sizeOfIn-1)*incr;
    out[iPos] = w1 * del1 + w2 * del2;
    if (dpchst(out[iPos], del2) <= 0)
    {
        out[iPos] = 0;
    }
    else if(dpchst(del1, del2) < 0)
    {
        double dmax = 3 * del2;
        if(fabs(out[iPos]) > fabs(dmax))
        {
            out[iPos] = dmax;
        }
    }
}

__kernel void splinCub_kernel(int _iPos, int sizeOfIn, int iSize, __global double* In, __global double* Z, __global double* Out,
                              __global double* Ad, __global double* Asd, __global double* Qdu, //double* ll,
                              int incr, SplineType spType)
{// natural, not_a_knot, clamped, periodic
    int iPosX = get_global_id(0);
    int iPosY = get_global_id(1);
    int iPos = _iPos;

    if(iPosX >= iSize || iPosY >= iSize)
    {
        return;
    }

    if(iPos == 0)
    {
        iPos = iPosY * sizeOfIn + iPosX;
    }

    __global double* z = Z + iPos;
    __global double* out = Out + iPos;
    __global double* ad = Ad + iPos;
    __global double* asd = Asd + iPos;
    __global double* qdu = Qdu + iPos;

    if(sizeOfIn == 2)
    {
        if (spType != CLAMPED)
        {
            out[0] = (z[incr] - z[0]) / (In[1] - In[0]);
            out[incr] = out[0];
        }
        return;
    }

    for(int i = 0; i < sizeOfIn-1; i++)
    {
        asd[i*incr] = 1 / (In[i+1] - In[i]);
        qdu[i*incr] = (z[(i+1) * incr] - z[i*incr]) * asd[i*incr] * asd[i*incr];
    }

    for(int i = 1; i < sizeOfIn-1; i++)
    {
        ad[i*incr] = 2 * (asd[(i-1)*incr] + asd[i*incr]);
        out[i*incr] = 3 * (qdu[(i-1)*incr] + qdu[i*incr]);
    }

    if(spType == NATURAL)
    {
        ad[0]  = 2 * asd[0];
        out[0] = 3 * qdu[0];
        ad[(sizeOfIn-1)*incr] = 2 * asd[(sizeOfIn-2)*incr];
        out[(sizeOfIn-1)*incr] = 3 * qdu[(sizeOfIn-2)*incr];
        triDiagLDLSolve(ad, asd, out, sizeOfIn, incr);
    }
    else if(spType == NOT_A_KNOT)
    {
        // s'''(x(2)-) = s'''(x(2)+)
        double r = asd[incr]/ asd[0];
        ad[0] = asd[0] / (1 + r);
        out[0] = ((3*r+2) * qdu[0] + r*qdu[incr]) / ((1+r)*(1+r));
        // s'''(x(n-1)-) = s'''(x(n-1)+)
        r = asd[(sizeOfIn-3)*incr] / asd[(sizeOfIn-2)*incr];
        ad[(sizeOfIn-1)*incr] = asd[(sizeOfIn-2)*incr] / (1 + r);
        out[(sizeOfIn-1) * incr] = ((3*r+2) * qdu[(sizeOfIn-2)*incr] + r*qdu[(sizeOfIn-3)*incr]) / ((1+r)*(1+r));
        triDiagLDLSolve(ad, asd, out, sizeOfIn, incr);
    }
    else if(spType == CLAMPED)
    {
        // d(1) and d(n) are already known
        out[incr] = out[incr] - out[0] * asd[0];
        out[(sizeOfIn-2)*incr] = out[(sizeOfIn-2)*incr] - out[(sizeOfIn-1)*incr] * asd[(sizeOfIn-2)*incr];
        triDiagLDLSolve(ad+incr, asd+incr, out+incr, sizeOfIn-2, incr);
    }
    else if(spType == PERIODIC)
    {
        ad[0] = 2 * (asd[0] + asd[(sizeOfIn-2)*incr]);
        out[0] = 3 * (qdu[0] + qdu[(sizeOfIn-2)*incr]);

        qdu[0] = asd[(sizeOfIn-2)*incr];
        for(int i = incr; i < (sizeOfIn-3)*incr; i+=incr)
        {
            qdu[i] = 0;
        }

        qdu[(sizeOfIn-3)*incr] = asd[(sizeOfIn-3)*incr];

        cyclicTriDiagLDLtSolve(ad, asd, qdu, out, sizeOfIn-1, incr);
        out[(sizeOfIn-1) * incr] = out[0];
    }
}

__kernel void coef_bicubic_kernel(__global double* X, __global double* Y, __global double* Z,
                                  int sizeOfX, int sizeOfY,
                                  __global double* P, __global double* Q, __global double* R, __global double* C)
{
    int iPosX = get_global_id(0);
    int iPosY = get_global_id(1);
    if(iPosX > sizeOfX-2 || iPosY > sizeOfY-2)
    {
        return;
    }

    int iPos    = iPosY * sizeOfX + iPosX;
    int iPosCp1 = (iPosY+1) * sizeOfX + iPosX;
    int iPosRp1 = iPosY * sizeOfX + iPosX + 1;

    double dy = 1 / (Y[iPosY+1] - Y[iPosY]);
    double dx = 1 / (X[iPosX+1] - X[iPosX]);

    double a = 0;
    double b = 0;
    double c = 0;
    double d = 0;

    C[computeCPos(0,0,iPosX,iPosY,sizeOfX)] = Z[iPos];
    C[computeCPos(1,0,iPosX,iPosY,sizeOfX)] = P[iPos];
    C[computeCPos(0,1,iPosX,iPosY,sizeOfX)] = Q[iPos];
    C[computeCPos(1,1,iPosX,iPosY,sizeOfX)] = R[iPos];

    a = (Z[iPosRp1] - Z[iPos]) * dx;
    C[computeCPos(2,0,iPosX,iPosY,sizeOfX)] = (3*a - 2*P[iPos] - P[iPosRp1])*dx;
    C[computeCPos(3,0,iPosX,iPosY,sizeOfX)] = (P[iPosRp1] + P[iPos] - 2*a)*dx*dx;

    a = (Z[iPosCp1] - Z[iPos]) * dy;
    C[computeCPos(0,2,iPosX,iPosY,sizeOfX)] = (3*a - 2*Q[iPos] - Q[iPosCp1])*dy;
    C[computeCPos(0,3,iPosX,iPosY,sizeOfX)] = (Q[iPosCp1] + Q[iPos] - 2*a)*dy*dy;

    a = (Q[iPosRp1] - Q[iPos]) * dx;
    C[computeCPos(2,1,iPosX,iPosY,sizeOfX)] = (3*a - 2*R[iPos] - R[iPosRp1])*dx;
    C[computeCPos(3,1,iPosX,iPosY,sizeOfX)] = (R[iPosRp1] + R[iPos] - 2*a)*dx*dx;

    a = (P[iPosCp1] - P[iPos]) * dy;
    C[computeCPos(1,2,iPosX,iPosY,sizeOfX)] = (3*a - 2*R[iPos] - R[iPosCp1])*dy;
    C[computeCPos(1,3,iPosX,iPosY,sizeOfX)] = (R[iPosCp1] + R[iPos] - 2*a)*dy*dy;

    a = (Z[iPosCp1+1] + Z[iPos] - Z[iPosRp1] - Z[iPosCp1]) * dx * dx *dy * dy
        - (P[iPosCp1] - P[iPos]) * dx *dy *dy
        - (Q[iPosRp1] - Q[iPos]) * dx *dx *dy
        + R[iPos] * dx *dy;
    b = (P[iPosCp1+1] + P[iPos] - P[iPosRp1] - P[iPosCp1]) * dx * dy * dy
        - (R[iPosRp1] - R[iPos]) * dx * dy;
    c = (Q[iPosCp1+1] + Q[iPos] - Q[iPosRp1] - Q[iPosCp1]) * dx * dx * dy
        - (R[iPosCp1] - R[iPos]) * dx * dy;
    d = (R[iPosCp1+1] + R[iPos] - R[iPosRp1] - R[iPosCp1]) * dx * dy;

    C[computeCPos(2,2,iPosX,iPosY,sizeOfX)] = 9 * a - 3 * b - 3 * c + d;
    C[computeCPos(2,3,iPosX,iPosY,sizeOfX)] = (-6 * a + 2 * b + 3 * c - d) * dy;
    C[computeCPos(3,2,iPosX,iPosY,sizeOfX)] = (-6 * a + 3 * b + 2 * c - d) * dx;
    C[computeCPos(3,3,iPosX,iPosY,sizeOfX)] = (4 * a - 2 * b - 2 * c + d) * dx * dy;
}
