/*
* Scilab ( http://www.scilab.org/ ) - This file is part of Scilab
* Copyright (C) Scilab Enterprises - 2014 - Cedric DELAMARRE
*
* This file must be used under the terms of the CeCILL.
* This source file is licensed as described in the file COPYING, which
* you should have received as part of this distribution.  The terms
* are also available at
* http://www.cecill.info/licences/Licence_CeCILL_V2-en.txt
*
*/

#include "config_gpu.h"
#ifdef WITH_Nvidia
#pragma OPENCL EXTENSION cl_khr_fp64: enable
#endif

__kernel void dsum_kernel(__global double* d, int elems, __global double* result, __local double* accumResult)
{
    int iGlobalPos  = get_global_id(0);
    int iLocalPos   = get_local_id(0);
    int iLocalSize  = get_local_size(0);
    int iGroupPos   = get_group_id(0);

    accumResult[iLocalPos] = 0.0;

    for(int i = iGlobalPos; i < elems; i += get_global_size(0))
            accumResult[iLocalPos] += d[i];

    barrier(CLK_LOCAL_MEM_FENCE);

    // if iLocalSize is odd.
    if(iLocalSize & 1 && iLocalPos == 0)
        accumResult[0] += accumResult[iLocalSize-1];

    for(int stride = iLocalSize >> 1; stride > 0; stride >>= 1)
    {
        if(iLocalPos < stride)
            accumResult[iLocalPos] += accumResult[stride + iLocalPos];

        barrier(CLK_LOCAL_MEM_FENCE);

        // if stride is odd, addition last element with the first
        // ie: matrix = 1 2 3 4 5 6 7
        // make 1 += 7
        if(stride & 1  && stride != 1 && iLocalPos == 0)
            accumResult[0] += accumResult[stride-1];

        barrier(CLK_LOCAL_MEM_FENCE);
    }

    if(iLocalPos == 0) result[iGroupPos] = accumResult[0];
}

__kernel void zsum_kernel(__global double2* d, int elems, __global double2* result, __local double2* accumResult)
{
    int iGlobalPos  = get_global_id(0);
    int iLocalPos   = get_local_id(0);
    int iLocalSize  = get_local_size(0);
    int iGroupPos   = get_group_id(0);

    accumResult[iLocalPos].x = 0.0;
    accumResult[iLocalPos].y = 0.0;

    for(int i = iGlobalPos; i < elems; i += get_global_size(0))
    {
        accumResult[iLocalPos].x += d[i].x;
        accumResult[iLocalPos].y += d[i].y;
    }

    barrier(CLK_LOCAL_MEM_FENCE);

    // if iLocalSize is odd.
    if(iLocalSize & 1 && iLocalPos == 0)
    {
        accumResult[0].x += accumResult[iLocalSize-1].x;
        accumResult[0].y += accumResult[iLocalSize-1].y;
    }

    for(int stride = iLocalSize >> 1; stride > 0; stride >>= 1)
    {
        if(iLocalPos < stride)
        {
            accumResult[iLocalPos].x += accumResult[stride + iLocalPos].x;
            accumResult[iLocalPos].y += accumResult[stride + iLocalPos].y;
        }

        barrier(CLK_LOCAL_MEM_FENCE);

        if(stride & 1 && stride != 1 && iLocalPos == 0)
        {
            accumResult[0].x += accumResult[stride-1].x;
            accumResult[0].y += accumResult[stride-1].y;
        }

        barrier(CLK_LOCAL_MEM_FENCE);
    }

    if(iLocalPos == 0)
    {
        result[iGroupPos].x = accumResult[0].x;
        result[iGroupPos].y = accumResult[0].y;
    }
}
